"""
    Proxy for Approval Process to simplify import statements and hide core package structure from client code.

    isort:skip_file
"""

from signoffs.core.process import (
    ApprovalsProcess,
    FsmApprovalsProcess,
    user_can_revoke_approval,
)
