{# template for displaying Django messages #}
    {% if messages %}
        {% for message in messages %}
            <div class="alert alert-{{ message.tags }}">
                <button type="button" class="close" data-dismiss="alert">×</button>
                {{ message }}
            </div>
        {% endfor %}
    {% endif %}
