{% load static %}

<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="utf-8">
        <meta http-equiv="X-UA-Compatible" content="IE=edge">
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <!-- The above 3 meta tags *must* come first in the head; any other head content must come *after* these tags -->
        <meta name="description" content="">
        <meta name="author" content="">

        <title>
            {% block title %}{% endblock title %} Signoffs Test App
        </title>

        <link type="text/css" rel="stylesheet" href="{% static 'css/signoffs.css' %}">
        {% block extra_head %}{% endblock extra_head %}
    </head>

    <body class="{% if user.is_authenticated %} logged-in{% endif %} {% block body_class %}{% endblock %}">
    {% block body %}
        <div class="container">

            {% include 'include/messages.html' %}

            {% block content %}{% endblock content %}

        </div> <!-- container -->
        {% block footer %}
            {% include 'include/footer.html' %}
        {% endblock footer %}

    {% endblock body %}

    {% block scripts %}{% endblock scripts %}

    {% block aux %}{% endblock aux %}

    </body>
</html>
